xquery version "3.1";

(:~
 : This module controls the serialization of any documents of the KMG.
 : Documents are given over by the TextGrid connector, prepared here, and given
 : over to the necessary functions for transformation/storage.
 :
 : @author Simon Sendler
 : @author Florian Barth
 : @version 1.1
 : @since 1.0
 :)
module namespace kmg-serial="http://mollenhauer-edition.de/ns/kmg-serial";

import module namespace kmg-register-transformation="https://mollenhauer-edition.de/ns/register-transformation" at "kmg-register-transformation.xqm";
import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace xml-utils="http://bdn-edition.de/portal/xml-utils" at "xml-utils.xqm ";
import module namespace bdn-odd="http://bdn-edition.de/ns/odd" at "schema.xqm";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace transformKMG="http://mollenhauer-edition.de/ns/transform" at "transform.xqm";
import module namespace kmg-util="http://mollenhauer-edition.de/ns/util" at "kmg-util.xqm";
import module namespace resolve="https://mollenhauer-edition.de/ns/resolve" at "util/resolve.xqm";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace sade="https://mollenhauer-edition.de/ns/sade";
declare namespace html = "http://www.w3.org/1999/xhtml";

declare boundary-space strip;

(:~
 : This function assures that any directories necessary to the serialization are
 : available.
 :)
declare function kmg-serial:directoryHelper() {
    let $resources := bdn-utils:assure-dir-available("resources")
    let $edition := bdn-utils:assure-dir-available("edition")
    let $edition := bdn-utils:assure-dir-available("lido")
    let $register := bdn-utils:assure-dir-available("register")
    let $aux := bdn-utils:assure-dir-available("aux")
    let $edition-aux := bdn-utils:assure-dir-available("edition-aux")
    let $tei := bdn-utils:assure-dir-available("tei")
    return
        true()
};

(:~
 : This function wraps the edition text in an article element used for indexing and
 : writes metadata to attributes.
 :)
declare function kmg-serial:build-article($content as node()*, $view as xs:string, $doc as node()*, $filename as xs:string)
as element(html:article)?
{
    element html:article {
        attribute data-view {
            if ($view = "digital" or $view = "commentary") then
                $view
            else if (kmg-util:getLeittextID($filename) = $view) then
                "main"
            else
                "other"
        },
        attribute data-wit {
            if ($view = "digital" or $view = "commentary") then
                ""
            else
                $view
        },
        attribute data-title {
            string-join(transformKMG:transform($doc//tei:title[@type="main"], $filename, "none", "", ""), " ")
            => normalize-space()
        },
        xml-utils:find-non-rdgs-and-non-quotes($content)
    }
};

(:~
 : This function injects the Series Statement, which is stored in a seperate
 : document in TextGrid and referenced in the edition, into edition documents.
 : This change is stored, and not done in memory during transformation, as
 : complete XML-documents should be offered for download.
 :
 : @param $doc A TEI-XML document
 : @param $filename The filename, used in logging
 : @return $doc A TEI-XML document with the full Series Statement
 :)
declare function kmg-serial:resolve-header-copies($doc as node()*, $filename as xs:string) {
    try {
            update replace $doc//tei:teiHeader//tei:seriesStmt
            with fn:doc($config:aux || "3qq81.xml")//*,
            util:log-system-out(util:system-time() || " " || $filename || ": Injection of seriesStmt successful."),
            $doc
        }
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Injection of seriesStmt failed. Some information might not be displayed correctly."),
                $doc
        },
    try {
        update replace $doc//tei:teiHeader//tei:projectDesc
        with fn:doc($config:aux || "3qq82.xml")//*,
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of projectDesc successful."),
        $doc
    }
    catch * {
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of projectDesc failed. Some information might not be displayed correctly."),
            $doc
    },
    try {
        update replace $doc//tei:teiHeader//tei:listPrefixDef
        with fn:doc($config:aux || "3wjxb.xml")//*,
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of listPrefixDef successful."),
        $doc
    }
    catch * {
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of seriesStmt listPrefixDef. Some information might not be displayed correctly."),
            $doc
    },
    try {
        update replace $doc//tei:teiHeader//tei:editoralDecl/tei:p
        with fn:doc($config:aux || "3qq83.xml")//*,
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of editorialDecl successful."),
        $doc
    }
    catch * {
        util:log-system-out(util:system-time() || " " || $filename || ": Injection of editorialDecl failed. Some information might not be displayed correctly."),
            $doc
    }
};

(:~
 : Transforms a given XML file to a HTML file for the single view for each version.
 :
 : @author Michelle Weidling
 : @author Max-Ferdinand Zeterberg
 : @author Simon Sendler
 : @see
 : @version 1.0
 : @param $file The XML file to be transformed
 : @param $filename
 : @return xs:string The path where the HTML is stored
 : :)
declare function kmg-serial:html-single($file as node()*,
$filename as xs:string*, $witLeittext as xs:string, $listWit as xs:string*) as xs:string*
{
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Start creating single views.")
    let $skippedWits := string-join($file//tei:witness[(fn:exists(@xml:id) and (@sade:singleview/string() = "false"))]/@xml:id/string(),", ")
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Skipping the following witnesses since singleview=false: " || $skippedWits)
    let $wits := $file//tei:witness[(fn:exists(@xml:id) and not(@sade:singleview/string() = "false"))]/@xml:id

    return
        for $wit in $wits

            let $log := util:log-system-out(util:system-time() || " " || $filename || ": Start creating single view for witness " || $wit || ".")
            let $content :=
                try {
                        transformKMG:transform($file, $filename, $wit, $witLeittext, $listWit)
                    }
                catch * {
                    util:log-system-out(util:system-time() || " " || $filename || ": Could not create single view for witness " || $wit || ".")
                }
            let $new-filename := substring-before($filename, ".") || "-" || $wit ||".html"
            let $editorials := element html:div { xml-utils:find-editorials($content) }
            let $content := kmg-serial:build-article($content, $wit, $file, $filename)
            let $new-filename-editorials := substring-before($filename, ".") || "-"|| $wit || "-editorials.html"

            return
                try {
                    xmldb:store($config:edition, $new-filename, $content),
                    xmldb:store($config:edition-aux, $new-filename-editorials, $editorials),
                    util:log-system-out(util:system-time() || " " || $filename || ": Finished single view for witness " || $wit || ".")}
                catch * {
                    util:log-system-out(util:system-time() || " " || $filename || ": Could not store single view:  " || $err:code || " " || $err:description)
                }
};

(:~
 : Transforms a given file to a HTML file for the digital view.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $file The current file in intermediate format
 : @param $filename
 : @return xs:string The path where the HTML is stored
 : :)
declare function kmg-serial:html-digital($file as node()*,
$filename as xs:string*, $witLeittext as xs:string, $listWit as xs:string*) as xs:string*
{
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Start creating digital view with Leittext " || $witLeittext)

    let $content :=
        try {
            transformKMG:transform($file, $filename, "digital", $witLeittext, $listWit)}
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Could not create digital view.")
        }
    let $rdgs-and-quotes := xml-utils:find-rdgs-and-quotes($content)
    let $variants-and-quotes := element html:div { $rdgs-and-quotes }
    let $editorials := element html:div { xml-utils:find-editorials($content) }
    let $content := kmg-serial:build-article($content, "digital", $file, $filename)
    let $wits := distinct-values($rdgs-and-quotes/@wits[string-length(.) = 1])

    let $new-filename := substring-before($filename, ".") || "-digital.html"
    let $new-filename-variants := substring-before($filename, ".") || "-digital-variants-and-quotes.html"
    let $new-filename-editorials := substring-before($filename, ".") || "-digital-editorials.html"

    return
        try {
            xmldb:store($config:edition, $new-filename, $content),
            xmldb:store($config:edition-aux, $new-filename-variants, $variants-and-quotes),
            xmldb:store($config:edition-aux, $new-filename-editorials, $editorials),
            util:log-system-out(util:system-time() || " " || $filename || ": Finished creating digital view.")}
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Could not create digital view.")
        }
};

declare function kmg-serial:html-commentary($file as node()*,
$filename as xs:string*, $witLeittext as xs:string, $listWit as xs:string*) as xs:string*
{
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Start creating textual commentary.")
    let $content :=
        try {
                transformKMG:transform($file, $filename, "commentary", $witLeittext, $listWit)
        }
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Could not transform textual commentary.")
        }
    let $new-filename := substring-before($filename, ".") || "-commentary.html"
    let $editorials := element html:div { xml-utils:find-editorials($content) }
    let $content := kmg-serial:build-article($content, "commentary", $file, $filename)
    let $new-filename-editorials := substring-before($filename, ".") || "-commentary-editorials.html"

    return
        try {
            xmldb:store($config:edition, $new-filename, $content),
            xmldb:store($config:edition-aux, $new-filename-editorials, $editorials),
            util:log-system-out(util:system-time() || " " || $filename || ": Finished creating textual commentary.")
        }
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Could not create textual commentary.")
        }
};

(:~
 : This function coordinates the serialization of edition documents and hands
 : the document over to any necessary functions.
 :
 : @param $doc A TEI-XML document
 :)
declare function kmg-serial:serialize-edition($uri) {
    let $move :=
        if (doc-available($config:sade-data || $uri)) then
            xmldb:move($config:sade-data, $config:tei, $uri)
        else ()
    let $doc := if (doc-available($config:sade-data || $uri)) then
                    doc($config:sade-data || $uri)
                else
                    doc($config:tei || $uri)
    let $filename := util:document-name($doc)
    let $doc := kmg-serial:resolve-header-copies($doc, $filename)
    let $update := resolve:includes($uri)
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Document upload complete. Commencing serialization.")
    let $listWitIDs := $doc//tei:teiHeader/tei:fileDesc/tei:sourceDesc/tei:listWit/tei:witness/@xml:id/string()
    let $file := $doc
    (:  MFZ, 2023-11-23: Why do we need this variable $file?! Why can't we continue with $doc only?  :)
    (:  MFZ, 2023-11-23: Diese Transformation neu eingefügt  :)
    (: let $doc := simplify:simplify-external($doc) :)
    let $listWit := ()
    let $listWit :=
        for $id in $listWitIDs return
            fn:insert-before($listWit, last(), $id)
    let $log :=
        if (count($listWit) > 0) then
            util:log-system-out(util:system-time() || " " || $filename || ": Identified " || count($listWit) || " witness(es).")
        else (
            util:log-system-out(util:system-time() || " " || $filename || ": Could not identify any wittnesses in tei:header."))

    let $witLeittext :=
        try {
            $file//tei:teiHeader/tei:fileDesc/tei:sourceDesc/tei:listWit/tei:witness/tei:desc[contains(@type, 'Leittext')]/../@xml:id/string()
        }
        catch * {
            util:log-system-out(util:system-time() || " " || $filename || ": Error: Could not identify witness with 'Leittext'.")
        }
    let $log := util:log-system-out(util:system-time() || " " || $filename || ": Identified witness '" || $witLeittext || "' as 'Leittext'.")

    let $single := kmg-serial:html-single($doc, $filename, $witLeittext, $listWit)
    let $digital :=
        if ( count($doc//tei:witness/@xml:id) gt 1 ) then
            kmg-serial:html-digital($doc, $filename, $witLeittext, $listWit)
        else
            util:log-system-out(util:system-time() || " " || $filename || ": Single witness text. Skipping digital view.")
    let $commentary := kmg-serial:html-commentary($doc, $filename, $witLeittext, $listWit)
    return
        util:log-system-out(util:system-time() || " " || $filename || ": Serialization complete.")
};

(:~
 : This function handles the storage of LIDO-XML.
 :
 : @param $uri The TextGrid-URI of the document
 :)
declare function kmg-serial:serialize-lido($uri) {
    xmldb:move($config:sade-data, $config:lido, $uri),
    util:log-system-out(util:system-time() || " " || $uri || ": Moved LIDO file.")
};

(:~
 : This function stores auxiliary files, such as the shared seriesStmt, in the
 : folder resources/aux
 :
 : @param $uri The document's TextGrid-URI
 :)
declare function kmg-serial:serialize-aux($uri as xs:string) {
    xmldb:move($config:sade-data, $config:aux, $uri)
};

declare function kmg-serial:update-indexes() {
    let $reindex-editions := xmldb:reindex($config:tei)
    let $reindex-register := xmldb:reindex($config:register)
    return
        true()
};

(:~
 : This function receives documents from the TextGrid connector, identifies them
 : and hands them over to their respective functions/modules for further
 : processing.
 :
 : @param $uri A TextGrid-URI of an XML document
 :)
declare function kmg-serial:serialize($uri) {
    let $prepare := kmg-serial:directoryHelper()

    return
        if ( $uri = "3r8zv.xml" ) then
            (bdn-odd:main(),
            kmg-serial:serialize-aux($uri))
        else if ( $uri = "3qq81.xml"
            or $uri = "3qq82.xml"
            or $uri = "3qq83.xml"
            or $uri = "3wjxb.xml"
            or $uri = "46ntv.xml") then
            kmg-serial:serialize-aux($uri)
        else
            let $doc :=
                if (doc-available($config:sade-data || $uri)) then
                    doc($config:sade-data || $uri)
                else
                    doc($config:tei || $uri)
            return
                if ( name($doc/*) = "lido:lido" ) then
                    (
                        kmg-serial:serialize-lido($uri),
                        kmg-serial:update-indexes(),
                        kmg-register-transformation:update-registers($uri)
                    )
                else
                    (
                        kmg-serial:serialize-edition($uri),
                        kmg-serial:update-indexes(),
                        kmg-register-transformation:update-registers($uri)
                    )
};